@@ -68,7 +68,7 @@ PAM_EXTERN int
 pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 		     int argc, const char **argv)
 {
-    int retval, i, citem=0, extitem=0, onerr=PAM_SERVICE_ERR, sense=2;
+    int retval, i, citem=0, extitem=0, onerr=PAM_SERVICE_ERR, sense=2, quiet=0;
     const void *void_citemp;
     const char *citemp;
     char *ifname=NULL;
@@ -108,16 +108,21 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 		onerr = PAM_SUCCESS;
 	    else if(!strcmp(myval,"fail"))
 		onerr = PAM_SERVICE_ERR;
-	    else
+	    else {
+	        if (ifname) free (ifname);
 		return PAM_SERVICE_ERR;
+	    }
 	else if(!strcmp(mybuf,"sense"))
 	    if(!strcmp(myval,"allow"))
 		sense=0;
 	    else if(!strcmp(myval,"deny"))
 		sense=1;
-	    else
+	    else {
+	        if (ifname) free (ifname);
 		return onerr;
+	    }
 	else if(!strcmp(mybuf,"file")) {
+	    if (ifname) free (ifname);
 	    ifname = (char *)malloc(strlen(myval)+1);
 	    if (!ifname)
 		return PAM_BUF_ERR;
@@ -150,6 +155,8 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 		    apply_type=APPLY_TYPE_USER;
 		    strncpy(apply_val,myval,sizeof(apply_val)-1);
 		}
+	    } else if (!strcmp(mybuf,"quiet")) {
+		quiet = 1;
 	    } else {
 		free(ifname);
 		pam_syslog(pamh,LOG_ERR, "Unknown option: %s",mybuf);
@@ -176,6 +183,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
               ) {
 	pam_syslog(pamh,LOG_ERR,
 		  "Invalid usage for apply= parameter");
+        free (ifname);
 	return onerr;
     }
 
@@ -231,6 +239,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
     retval = pam_get_item(pamh,citem,&void_citemp);
     citemp = void_citemp;
     if(retval != PAM_SUCCESS) {
+	free(ifname);
 	return onerr;
     }
     if((citem == PAM_USER) && !citemp) {
@@ -393,8 +402,9 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 #endif
 	(void) pam_get_item(pamh, PAM_SERVICE, &service);
 	(void) pam_get_user(pamh, &user_name, NULL);
-	pam_syslog (pamh, LOG_ALERT, "Refused user %s for service %s",
-		    user_name, (const char *)service);
+	if (!quiet)
+	    pam_syslog (pamh, LOG_ALERT, "Refused user %s for service %s",
+	                user_name, (const char *)service);
 	return PAM_AUTH_ERR;
     }
 }
